/*@/// Intro */
// Switch on string

// (c) 2011..2020 Sigrid/Jasper Neumann
// www.sirrida.de / programming.sirrida.de
// E-Mail: info@sirrida.de

// Granted to the public domain
// First version: 2013-02-23
// Last change: 2013-04-12

// Hash via optimized map (STL strings).
// g++
/*@\\\+E24C*/

// Options:
//   DEBUG_CASE_DUPE: Check for dupes at runtime: Output an error message.
//   DEBUG_CASE_STATISTICS: Dump a map statistics

#ifndef unit__case_ms
#define unit__case_ms

#include "general.h"
#include "modula.h"

#include <iostream>
#include <string>
using namespace std;


mycall t_int floor_power_2(t_int x);
mycall t_int ceil_power_2(t_int x);

/*@/// typedef struct tr_string_entry */
typedef struct tr_string_entry {
  struct tr_string_entry* next;
  string s;
  t_int i;
  } tr_string_entry;
/*@\\\0000000308+0D93*/
/*@/// class tr_string_map */
class tr_string_map {
public:
  mycall t_int (*hash)(const string& s);
  tr_string_entry** map;
  t_int nr_slots;
  t_int mask;
  volatile t_bool finalized;

  tr_string_map();
  t_bool init();
  void finalize();
  t_int find(const string& s, t_int not_found);
  };
/*@\\\000000070C+7C9F*/

#define LBL(x) (__LINE__-(t_int)(_o_)+2-(x))
  // The fiddling with _o_ shifts the range's lower bound to 0.
/*@/// #define SWITCH(s) */
// On first run let _l_=0 and collect the CASE descriptors,
// then with _l_=1 apply string_map_finalize to determine
// the best suiting hash function, build a hash table,
// and thereby reverse the linking order of each chain
// which effectively links the first appearing CASE entries first.
// On this and all later runs (_l_=2) calculate the hash of the given string
// and loop through all strings of the corresponding hash slot.
// When the string is found, dispatch with switch to the given line.
#define SWITCH(s) \
{{ \
  enum {_o_ = __LINE__}; \
  static tr_string_map _m_; \
  t_int _i_; \
  t_int _l_; \
  for (_l_=(_m_.finalized?2:0); _l_<=2; ++_l_) { \
    if (_l_ == 2) { \
      _i_ = _m_.find((s), LBL(1)); \
      } \
    else if (_l_ == 0) { \
      if (_m_.init()) { \
        _l_ = 1; \
        continue; \
        } \
      _i_ = LBL(2); \
      } \
    else { \
      _m_.finalize(); \
      continue; \
      } \
    switch (_i_) { \
      case LBL(2): if (false) {
/*@\\\0000001F20+FE14*/
/*@/// #define CASE(s) */
// Define one struct per CASE and link it before the previous one.
#define CASE(s) \
        break; \
        } \
      if (_l_ == 0) { \
        static tr_string_entry _e_ = {0, (s), LBL(0)}; \
        _e_.next = _m_.map[0]; \
        _m_.map[0] = &_e_; \
        } \
      else \
        case LBL(0): {
/*@\\\0000000512+4AB6*/
/*@/// #define DEFAULT */
#define DEFAULT \
        break; \
        } \
      if (_l_ != 0) \
        default: {
/*@\\\0000000513+FD63*/

/*@/// #define CASE_MULTI0 */
#define CASE_MULTI0 \
        break; \
        } \
      { \
        enum {_x_ = LBL(0)};
/*@\\\+08B4*/
/*@/// #define CASE_MULTI(s) */
#define CASE_MULTI(s) \
        break; \
        } \
      { \
        enum {_x_ = LBL(0)}; \
        if (_l_ == 0) { \
          static tr_string_entry _e_ = {0, (s), (t_int)(_x_)}; \
          _e_.next = _m_.map[0]; \
          _m_.map[0] = &_e_; \
          }
/*@\\\0000000A0C+CA00*/
/*@/// #define CASE_OR(s) */
#define CASE_OR(s) \
        if (_l_ == 0) { \
          static tr_string_entry _e_ = {0, (s), (t_int)(_x_)}; \
          _e_.next = _m_.map[0]; \
          _m_.map[0] = &_e_; \
          }
/*@\\\000000060C+95EE*/
/*@/// #define CASE_DO */
#define CASE_DO \
        case (t_int)(_x_): {} \
        } \
      if (_l_ != 0) {
/*@\\\0000000416+7A7C*/

#endif

// eof.
/*@\\\0000002508+F944*/
